@@ -1,15 +1,15 @@
 import datetime
-import six
 import warnings
 
+import six
+
+from .. import exceptions
+from .base import InstanceResource, ListResource
+
 if six.PY3:
     from io import IOBase as file
 
 
-from .base import InstanceResource, ListResource
-from .. import exceptions
-
-
 class CommentableResource(InstanceResource):
     """
     CommentableResource base class
@@ -267,11 +267,133 @@ class UserStoryAttachments(Attachments):
     instance = UserStoryAttachment
 
 
+class EpicAttachment(Attachment):
+    """
+    EpicAttachment class
+    """
+    endpoint = 'epics/attachments'
+
+
+class EpicAttachments(Attachments):
+    """
+    EpicAttachments factory class
+    """
+    instance = EpicAttachment
+
+
+class Epic(CustomAttributeResource, CommentableResource):
+    """
+    Epic model
+
+    :param assigned_to: assigned to property of the :class:`Epic`
+    :param blocked_note: blocked note of the :class:`Epic`
+    :param description: description of of the :class:`Epic`
+    :param is_blocked: is blocked property of the :class:`Epic`
+    :param is_closed: is closed property of the :class:`Epic`
+    :param color: the color of the :class:`Epic`
+    :param project: the project of the :class:`TaskStatus`
+    :param subject: subject of the :class:`TaskStatus`
+    :param tags: tags of the :class:`TaskStatus`
+    :param watchers: watchers of the :class:`TaskStatus`
+    :param version: version of the :class:`Epic`
+    """
+
+    endpoint = 'epics'
+
+    repr_attribute = 'subject'
+
+    allowed_params = [
+        'assigned_to', 'blocked_note', 'description',
+        'is_blocked', 'is_closed', 'color', 'project',
+        'subject', 'tags', 'watchers', 'version'
+    ]
+
+    def list_user_stories(self, **queryparams):
+        """
+        Returns the :class:`UserStory` list of the project.
+        """
+        return UserStories(self.requester).list(epic=self.id, **queryparams)
+
+    def list_attachments(self):
+        """
+        Get a list of :class:`EpicAttachment`.
+        """
+        return EpicAttachments(self.requester).list(object_id=self.id)
+
+    def attach(self, attached_file, **attrs):
+        """
+        Attach a file to the :class:`Epic`
+
+        :param attached_file: file path to attach
+        :param attrs: optional attributes for the attached file
+        """
+        return EpicAttachments(self.requester).create(
+            self.project, self.id,
+            attached_file, **attrs
+        )
+
+
+class Epics(ListResource):
+    """
+    Epics factory class
+    """
+    instance = Epic
+
+    def create(self, project, subject, **attrs):
+        """
+        Create a new :class:`Epic`.
+
+        :param project: :class:`Project` id
+        :param subject: subject of the :class:`Epic`
+        :param attrs: optional attributes of the :class:`Epic`
+        """
+        attrs.update({'project': project, 'subject': subject})
+        return self._new_resource(payload=attrs)
+
+
+class EpicStatus(InstanceResource):
+    """
+    Taiga Epic Status model
+
+    :param color: the color of the :class:`EpicStatus`
+    :param is_closed: closed property of the :class:`EpicStatus`
+    :param name: The name of the :class:`EpicStatus`
+    :param order: order of the :class:`EpicStatus`
+    :param project: the Taiga project of the :class:`EpicStatus`
+    :param slug: the slug of the :class:`EpicStatus`
+    """
+
+    repr_attribute = 'subject'
+
+    endpoint = 'epic-statuses'
+
+    allowed_params = [
+        'color', 'is_closed', 'name', 'order', 'project', 'slug`'
+    ]
+
+
+class EpicStatuses(ListResource):
+
+    instance = EpicStatus
+
+    def create(self, project, name, **attrs):
+        """
+        Create a new :class:`EpicStatus`.
+
+        :param project: :class:`Project` id
+        :param name: name of the :class:`EpicStatus`
+        :param attrs: optional attributes of the :class:`EpicStatus`
+        """
+        attrs.update({'project': project, 'name': name})
+        return self._new_resource(payload=attrs)
+
+
 class UserStory(CustomAttributeResource, CommentableResource):
     """
     User Story model
 
     :param assigned_to: assigned to of the :class:`UserStory`
+    :param assigned_users: additional users assigned to of the :class:`UserStory`
     :param backlog_order: backlog order of the :class:`UserStory`
     :param blocked_note: blocked note of the :class:`UserStory`
     :param version: version of the :class:`UserStory`
@@ -288,16 +410,22 @@ class UserStory(CustomAttributeResource, CommentableResource):
     :param tags: tags of the :class:`UserStory`
     :param team_requirement: team requirement of the :class:`UserStory`
     :param watchers: watchers of the :class:`UserStory`
+    :param due_date: :class:`UserStory` due date
+    :param generated_from_issue: :class:`UserStory` parent issue
+    :param generated_from_task: :class:`UserStory` parent task
     """
     endpoint = 'userstories'
 
     repr_attribute = 'subject'
+    element_type = 'User Story'
+    element_shortcut = 'us'
 
     allowed_params = [
-        'assigned_to', 'backlog_order', 'blocked_note', 'version',
+        'assigned_to', 'assigned_users', 'backlog_order', 'blocked_note', 'version',
         'client_requirement', 'description', 'is_blocked', 'is_closed',
         'kanban_order', 'milestone', 'points', 'project', 'sprint_order',
-        'status', 'subject', 'tags', 'team_requirement', 'watchers'
+        'status', 'subject', 'tags', 'team_requirement', 'watchers', 'due_date',
+        'generated_from_issue', 'generated_from_task',
     ]
 
     def add_task(self, subject, status, **attrs):
@@ -587,11 +715,14 @@ class Task(CustomAttributeResource, CommentableResource):
     :param is_iocaine: the is iocaine of the :class:`TaskStatus`
     :param external_reference: external reference of the :class:`TaskStatus`
     :param watchers: watchers of the :class:`TaskStatus`
+    :param due_date: :class:`Task` due date
     """
 
     endpoint = 'tasks'
 
     repr_attribute = 'subject'
+    element_type = 'Task'
+    element_shortcut = 'task'
 
     allowed_params = [
         'assigned_to', 'blocked_note', 'description', 'version',
@@ -738,12 +869,14 @@ class Issue(CustomAttributeResource, CommentableResource):
     :param subject: subject of the issue
     :param tags: array of tags
     :param watchers: array of watchers id
-
+    :param due_date: :class:`Issue` due date
     """
 
     endpoint = 'issues'
 
     repr_attribute = 'subject'
+    element_type = 'Issue'
+    element_shortcut = 'issue'
 
     allowed_params = [
         'assigned_to', 'blocked_note', 'description', 'version',
@@ -977,9 +1110,27 @@ class Project(InstanceResource):
         'severities': Severities,
         'roles': Roles,
         'points': Points,
-        'us_statuses': UserStoryStatuses
+        'us_statuses': UserStoryStatuses,
+        'milestones': Milestones
     }
 
+    def get_item_by_ref(self, ref):
+        response = self.requester.get(
+            '/resolver?project={project_id}&ref={task_ref}',
+            task_ref=ref,
+            project_id=self.slug
+        )
+        response_json = response.json()
+
+        if response_json and 'task' in response_json:
+            return self.get_task_by_ref(ref)
+        elif response_json and 'us' in response_json:
+            return self.get_userstory_by_ref(ref)
+        elif response_json and 'issue' in response_json:
+            return self.get_issue_by_ref(ref)
+        else:
+            return None
+
     def get_task_by_ref(self, ref):
         """
         Get a :class:`Task` by ref.
@@ -994,6 +1145,20 @@ class Project(InstanceResource):
         )
         return Task.parse(self.requester, response.json())
 
+    def get_epic_by_ref(self, ref):
+        """
+        Get a :class:`Epic` by ref.
+
+        :param ref: :class:`Epic` reference
+        """
+        response = self.requester.get(
+            '/{endpoint}/by_ref?ref={ep_ref}&project={project_id}',
+            endpoint=Epic.endpoint,
+            ep_ref=ref,
+            project_id=self.id
+        )
+        return Epic.parse(self.requester, response.json())
+
     def get_userstory_by_ref(self, ref):
         """
         Get a :class:`UserStory` by ref.
@@ -1220,11 +1385,11 @@ class Project(InstanceResource):
             estimated_finish, **attrs
         )
 
-    def list_milestones(self):
+    def list_milestones(self, **queryparams):
         """
         Get the list of :class:`Milestone` resources for the project.
         """
-        return Milestones(self.requester).list(project=self.id)
+        return Milestones(self.requester).list(project=self.id, **queryparams)
 
     def add_point(self, name, value, **attrs):
         """
@@ -1242,6 +1407,23 @@ class Project(InstanceResource):
         """
         return Points(self.requester).list(project=self.id)
 
+    def add_epic(self, subject, **attrs):
+        """
+        Adds a :class:`UserStory` and returns a :class:`UserStory` resource.
+
+        :param subject: subject of the :class:`UserStory`
+        :param attrs: other :class:`UserStory` attributes
+        """
+        return Epics(self.requester).create(
+            self.id, subject, **attrs
+        )
+
+    def list_epics(self):
+        """
+        Get the list of :class:`Epic` resources for the project.
+        """
+        return Epics(self.requester).list(project=self.id)
+
     def add_task_status(self, name, **attrs):
         """
         Add a Task status to the project and returns a
@@ -1489,6 +1671,30 @@ class Project(InstanceResource):
         """
         return Webhooks(self.requester).list(project=self.id)
 
+    def add_tag(self, tag, color=None):
+        """
+        Add a new tag and return a response object.
+
+        :param tag: name of the tag
+        :param color: optional color of the tag
+        """
+        attrs = {'tag': tag}
+        if color:
+            attrs['color'] = color
+        response = self.requester.post(
+            "/{}/{}/create_tag".format(self.endpoint, self.id), payload=attrs
+        )
+        return response
+
+    def list_tags(self):
+        """
+        Get the list of tags for the project.
+        """
+        response = self.requester.get(
+            "/{}/{}/tags_colors".format(self.endpoint, self.id)
+        )
+        return response.json()
+
 
 class Projects(ListResource):
     """
@@ -1654,6 +1860,7 @@ class History(InstanceResource):
         self.task = HistoryTask(self.requester)
         self.user_story = HistoryUserStory(self.requester)
         self.wiki = HistoryWiki(self.requester)
+        self.epic = HistoryEpic(self.requester)
 
 
 class HistoryEntity(object):
@@ -1673,7 +1880,8 @@ class HistoryEntity(object):
         """
         response = self.requester.get(
             '/{endpoint}/{entity}/{id}',
-            endpoint=self.endpoint, entity=self.entity, id=resource_id
+            endpoint=self.endpoint, entity=self.entity, id=resource_id,
+            paginate=False
         )
         return response.json()
 
@@ -1713,6 +1921,15 @@ class HistoryIssue(HistoryEntity):
         self.entity = 'issue'
 
 
+class HistoryEpic(HistoryEntity):
+    """
+    HistoryEpic model
+    """
+    def __init__(self, *args, **kwargs):
+        super(type(self), self).__init__(*args, **kwargs)
+        self.entity = 'epic'
+
+
 class HistoryTask(HistoryEntity):
     """
     HistoryTask model
